import pyaudio
import wave
import keyboard
import threading
import subprocess
import serial
import time
import queue
import sys

#CODEC2 mode              1600    700      700C
#bits pet frame           64      28       28
#min frame ms             40      80       80
#min frame size byte      8       7        7
#target frame bytes       <57     <57      <57
#min frame in frame       7       8        8
#frame size taken         56      56       56 
#frame size ms            280     640      640
#RAW input Rate frames/s  8000    8000     8000
#RAW input frames         2240    5120     5120
#input frames size byte   4480    10240    10240  <-dec process buff is less????
#coded to size in bytes           64
#8960
#4480

PLAY_DELAY = 1.0  #seconds
CHUNK = 4480#5120 #in 8000hz 16b sample size: 8 samples per 1 ms or 280 per 35ms eq in 4480b OR 2240 samples per 280ms for perfect codec2 enc eq in 2240*16bits/sample = 35840 bits   
FORMAT = pyaudio.paInt16
CHANNELS = 1
RATE = 8000 #44100
OUTPUT_2_WAV     = True
WAV_RECIEVE_NAME = '_out.wav'
WAV_SEND_NAME    = '_mic.wav'
#ENC_CALL_PATH = '/mnt/c/devel/lorawan/codec2-0.8/build_linux/src/c2enc'
#DEC_CALL_PATH = '/mnt/c/devel/lorawan/codec2-0.8/build_linux/src/c2dec'
ENC_CALL_PATH = '/home/snow/codec2/0.8/build1/src/c2enc'
DEC_CALL_PATH = '/home/snow/codec2/0.8/build1/src/c2dec'
ENC_CALL_STR = '' #'bash -c "/home/snow/codec2/0.8/build1/src/c2enc 700C - -"'
DEC_CALL_STR = '' #'bash -c "/home/snow/codec2/0.8/build1/src/c2dec 700C - -"'
COM_PORT = 'COM3'
extra_bytes_for_coded_data = bytearray()

stream = pyaudio.Stream
p = pyaudio.PyAudio
serial_port = serial.Serial()
packet_counter = 0x00
recording_flag = False

q_2enc = queue.Queue()
q_enc = queue.Queue()
q_2dec = queue.Queue()
q_dec = queue.Queue()

class ThreadRec(threading.Thread):
  def __init__(self):
    super(ThreadRec, self).__init__()
    self.stoprequest = threading.Event()
    self.recording = threading.Event()

  def run(self):
    if OUTPUT_2_WAV:
      wf = wave.open(WAV_SEND_NAME, 'wb')
      wf.setnchannels(CHANNELS)
      wf.setsampwidth(p.get_sample_size(FORMAT))
      wf.setframerate(RATE)
    print('Record thread started')
    while not self.stoprequest.isSet():
      while self.recording.isSet():
        data = stream.read(CHUNK)#, False)
        #print('REC recorded bytes per CHUNK ',len(data))
        if not self.recording.isSet():
          stream.stop_stream()
        if OUTPUT_2_WAV : wf.writeframes(data)    
        q_2enc.put(data)
        
        #in 8000hz 16b sample size: 8 samples per 1 ms or 280 per 35ms eq in 4480b   
        #print(len(data)*8) #exactly 4480 bits per buffer got from mic        
      time.sleep(0.25)
    if OUTPUT_2_WAV: wf.close()
    print("Recording Thread stopped")

  def join(self, timeout=None):
    self.stoprequest.set()
    super(ThreadRec, self).join(timeout)

  def start_rec(self):
    stream.start_stream()
    self.recording.set()

  def stop_rec(self):
    self.recording.clear()

rec_thread = ThreadRec()

class ThreadEnc(threading.Thread):
  def __init__(self):
    super(ThreadEnc, self).__init__()
    self.stoprequest = threading.Event()
    self.p1 = subprocess.Popen(ENC_CALL_STR, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    
  def run(self):
    #in 1600 b/s 35 ms eq 56b
    #we need packet up to 58 bytes
    #for folowing 40ms frames for codec2
    #200 bytes/s > 280ms ~ 56 bytes

    #codec.c if  
    #codec2_bits_per_frame():
    #(c2->mode == CODEC2_MODE_1600)
    #return 64;
    #so 8bytes per frame, maximum number of frames per 58bytes packet fpr serial is 7 ~eq 56bytes per packet
    print('Encode thread started')
    #buff = bytearray()
    while not self.stoprequest.isSet():
      try: 
        buff = q_2enc.get(True,1.0) #get msg with timeout 1.0 sec
        q_2enc.task_done()
        print('{0:.3f} ENC started enc'.format(time.time()))
        #self.p1.stdin.write(buff[0:5120])
        self.p1.stdin.write(buff)
        self.p1.stdin.flush()
        #print('ENC stdout PEEK size bytes {}'.format(len(self.p1.stdout.peek())))
        #time.sleep(1)
        #print('ENC left in buffer (peek) ', len(self.p1.stdout.peek()))

        #i=0
        #while i < 1000:
        #  i +=1
        #  print('number ', i, ' ', self.p1.stdout.read(1))
        
        if extra_bytes_for_coded_data :
          tmp = self.p1.stdout.read(54)
          out = tmp + extra_bytes_for_coded_data
        else:
          out = self.p1.stdout.read(56)
        
        #print('ENC left in buffer afer coding ', len(self.p1.stdout.peek()))
        #self.p1.stdin.write(buff[5120:10240])
        #self.p1.stdin.flush()
        #time.sleep(1)
        #print('ENC stdout PEEK size bytes {}'.format(len(self.p1.stdout.peek())))
        #out1 = self.p1.stdout.read(28) #precise read size required
        #print('left in buffer afer coding2 ', len(self.p1.stdout.peek()))
        print('{0:.3f} ENC ecoded'.format(time.time()))
        q_enc.put(out)
        print('{:0.3f} ENC  to send queue {}'.format(time.time(),out.hex()))
        #print(len(out+out1))
        #print('ENC {} put to queue'.format(time.time()))
      except:
        continue
    self.p1.communicate()
    print('Encoding thread stopped.')

  def join(self, timeout=None):
    self.stoprequest.set()
    super(ThreadEnc, self).join(timeout)

enc_thread = ThreadEnc

class ThreadSerial(threading.Thread):
  def __init__(self):
    super(ThreadSerial, self).__init__()
    self.stoprequest = threading.Event()
    
  def run(self):
    drop_counter_delay_cycle = 0
    global packet_counter
    total_send = 0
    total_recieve = 0
    print('Serial thread started')
    data_rcv = bytearray()
    drop_buff = bytearray()
    last_byte_time = time.time()
    #first_byte = True
    while not self.stoprequest.isSet():
      #if we got something to read - READ IT, then if we got something to write - WRITE IT
      if serial_port.in_waiting >= 1: #if we got all 61 bytes to read
        b = serial_port.read(1)
        #if from the last time we got a byte passed more than 2*CHUNKS(ms) ~ 1000(ms) then drop buffers data_rcv and drop_buff
        if (last_byte_time - time.time()) > 1.0:
          drop_buff.extend(data_rcv)
          data_rcv.clear()
          print('{:.3f} RCV DROP bytes {} size {}'.format(time.time(), drop_buff.hex(), len(drop_buff)))
          drop_buff.clear()
        last_byte_time = time.time()
        data_rcv.extend(b)
        if len(data_rcv) >= 61 :
          #print('{:.3f} RCV got packet {} size {}'.format(time.time(), data_rcv.hex(), len(data_rcv)))
          #check packet first bits if not FFFF17 wait for beginning
          if data_rcv[0:3].hex() == 'ffff17':
            if len(drop_buff) > 0:
              print('{:.3f} RCV DROP bytes {} size {}'.format(time.time(), drop_buff.hex(), len(drop_buff)))
              drop_buff.clear()
            total_recieve += 1
            print('{:.3f} RCV recieved {} totap packets recieved: {} '.format(time.time(), data_rcv.hex(), total_recieve))
            if extra_bytes_for_coded_data :
              q_2dec.put(data_rcv[3:57])
            else:
              q_2dec.put(data_rcv[3:59])
            #q_2dec.put(data_rcv[3:59])
            del data_rcv[0:62]
          else:
            #if first bytes are not FFFF17 drop the bytes one by one from the buffer
            drop_buff.extend(bytes([data_rcv.pop(0)]))
            #print('popped byte ', hex(data_rcv.pop(0)))
      else:
        try: 
          buff = q_enc.get(True,0.3) #get msg with timeout 0.3 sec
          q_enc.task_done()
          #print(q_2enc.empty())
          #to send data we need 58 bytes so we have 2 free bytes at the end, let us put packet number in them
          packet_counter = packet_counter + 0x01
          data = bytearray()
          data.extend([0xff,0xff,0x17]) #adress of reciever device 2 bytes + 1 byte chanel number
          data.extend(buff) #56 bytes of data
          data.extend([0x00,packet_counter]) # one empty byte and packet number 
          #if packet_counter < 7:
          #  del data[5:10]
          print('{:.3f} SND for send {} bytes {}'.format(time.time(),data.hex(), len(data)))
          #if packet_counter ==2:
          #  data[0] = 0x00
          bytes_written = serial_port.write(data)
          #print('{:.3f} SND bytes send {} packet number {}'.format(time.time(),bytes_written, packet_counter))
          #print('SND {}'.format(data.hex()))
          #drop_counter_delay_cycle = 0
          continue
        except:
          #print('packet counter is : {}'.format(packet_counter))
          #drop_counter_delay_cycle += 1
          #if drop_counter_delay_cycle > 7: #7*0.3=2.1 sec no packets drop counter
          #  packet_counter = 0x00
          #  drop_counter_delay_cycle = 0
          continue
    print("Serial Thread stopped.")

  def join(self, timeout=None):
    self.stoprequest.set()
    super(ThreadSerial, self).join(timeout)

serial_thread = ThreadSerial()

class ThreadDecode(threading.Thread):
  def __init__(self):
    super(ThreadDecode, self).__init__()
    self.stoprequest = threading.Event()
    self.p1 = subprocess.Popen(DEC_CALL_STR, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    
  def run(self):
    if OUTPUT_2_WAV:
      wdf = wave.open(WAV_RECIEVE_NAME, 'wb')
      wdf.setnchannels(CHANNELS)
      wdf.setsampwidth(p.get_sample_size(FORMAT))
      wdf.setframerate(RATE)
    print('Decode thread started')
    while not self.stoprequest.isSet():
      #if we got something to to decode - read it from buffer and send to decoder, then write to wav
      try: 
        buff = q_2dec.get(True,1.0) #get msg with timeout 1.0 sec
        #print('got buff in dec')
        q_2dec.task_done()
        print('{:.3f} DEC started dec    {}'.format(time.time(), buff.hex()))
        self.p1.stdin.write(buff)
        self.p1.stdin.flush()
        #time.sleep(1)
        #i = 1
        #while i < 11000:
          #print(self.p1.stdout.peek(i))
        #  print(i)
        #  i += 1
        #print('STDOUT PEEK size bytes {}'.format(len(self.p1.stdout.peek())))
        data = self.p1.stdout.read(CHUNK*2) #precise read size required
        #print('DEC peek size after buff read bytes {}'.format(len(self.p1.stdout.peek())))
        print('{0:.3f} DEC finished dec'.format(time.time()))
        q_dec.put(data)
        if OUTPUT_2_WAV: wdf.writeframes(data)
      except:
        continue
    if OUTPUT_2_WAV: wdf.close()
    self.p1.communicate()
    print('Decoding thread stopped.')

  def join(self, timeout=None):
    self.stoprequest.set()
    super(ThreadDecode, self).join(timeout)

decode_thread = ThreadDecode

class ThreadPlay(threading.Thread):
  def __init__(self):
    super(ThreadPlay, self).__init__()
    self.stoprequest = threading.Event()
    
  def run(self):
    to_play = False
    global PLAY_DELAY #= 3.0  #seconds
    data = bytearray()
    first = True
    p3 = pyaudio.PyAudio() #using global p*??
    stream3 = p3.open(format=FORMAT,
      channels=CHANNELS,
      frames_per_buffer=CHUNK,
      rate=RATE,
      output=True)
    print('Play thread started')
    while not self.stoprequest.isSet():
      try:
        data = q_dec.get(True,1)
        q_dec.task_done()
        #print('{} got new chunk from queue'.format(time.time()))
        if first:
          print('PLA playing will start in {} sec'.format(PLAY_DELAY))
          time.sleep(PLAY_DELAY)
          first = False
        #print('{} frames available to write {}'.format(time.time(), stream3.get_write_available()))
        stream3.write(bytes(data))
        #print('{} sent to stream. frames av {} '.format(time.time(),stream3.get_write_available()))
        #continue
      except Exception:
        #print('{} no sound in play q'.format(time.time()))
        if not first:
          print('PLA stopped playing')
        first = True
        continue
    stream3.stop_stream()
    stream3.close()
    p3.terminate()
    print('Playing thread stopped.')

  def join(self, timeout=None):
    self.stoprequest.set()
    super(ThreadPlay, self).join(timeout)

play_thread = ThreadPlay()

#INITIALIZE EVERYTHING
def initAll():
  global p
  global stream
  global serial_port
  global decode_thread, enc_thread
  print('INIT Initializing components')
  #read_conf_file('config.txt')
  p = pyaudio.PyAudio()
  stream = p.open(format=FORMAT,
    channels=CHANNELS,
    rate=RATE,
    input=True,
    frames_per_buffer=CHUNK)
  rec_thread.start()
  serial_port.baudrate = 9600
  serial_port.port = COM_PORT#'COM3'
  serial_port.open()
  serial_thread.start()
  decode_thread = ThreadDecode()
  decode_thread.start()
  enc_thread = ThreadEnc()
  enc_thread.start()
  play_thread.start()
  print('INIT done')

#DESTROY EVETHING
def destAll():
  global p
  global stream
  print('DEST Freeing all interfaces')
  rec_thread.join()
  stream.stop_stream()
  stream.close()
  p.terminate()
  serial_thread.join()
  serial_port.close()
  decode_thread.join()
  enc_thread.join()
  play_thread.join()
  print('DEST everything terminated...')

def on_press(key):
  global p#, stream 
  global recording_flag
  global rec_thread
  global serial_port
  #print(key.scan_code)
  if key.scan_code == 42:
    #start tranmition to COM
    if not recording_flag and key.event_type == 'down':
      print('REC starting recording...')
      recording_flag = True
      rec_thread.start_rec()
    elif recording_flag and key.event_type == 'up':
      rec_thread.stop_rec()
      recording_flag = False
      print('REC stopped recording...')
  if (key.name == 's' and key.event_type == 'down'):
    #show state
    host_mme = p.get_default_host_api_info()
    print('MIC %s' % p.get_device_info_by_host_api_device_index(host_mme['index'], host_mme['defaultInputDevice']))
    print('SER ', serial_port)
    #add all other configs
    print('PLAY_DELAY ',PLAY_DELAY)  #seconds
    print('CHUNK ', CHUNK)
    print('FORMAT ',FORMAT)
    print('CHANNELS ',CHANNELS)
    print('RATE ',RATE)
    print('OUTPUT_2_WAV ',OUTPUT_2_WAV)
    print('WAV_RECIEVE_NAME ',WAV_RECIEVE_NAME)
    print('WAV_SEND_NAME ',WAV_SEND_NAME)
    print('ENC_CALL_STR ',ENC_CALL_STR)
    print('DEC_CALL_STR ',DEC_CALL_STR)
    print('COM_PORT ',COM_PORT)
  if (key.name == 'i' and key.event_type == 'down'):
    print('SER bytes waiting to read: {}'.format(serial_port.in_waiting))
  if (key.name == 'd' and key.event_type == 'down'):
    b = serial_port.read(serial_port.in_waiting);
    print('SER droped buffer to {} readin {} bytes to dev/nul'.format(serial_port.in_waiting, len(b)))
    print('SER dropped buffer:')
    print(b)
  if (key.name == 'esc'):
    keyboard.unhook(on_press)
    print('Press ESC to exit')
  
def main():
  initAll()
  keyboard.hook(on_press, suppress=False)
  keyboard.wait('esc')
  destAll()
  
def print_globals():
  print('CHUNK {}'.format(CHUNK))
  print('ENC_CALL_STR {}'.format(ENC_CALL_STR))
  print('DEC_CALL_STR {}'.format(DEC_CALL_STR))
  print('COM_PORT {}'.format(COM_PORT))
  print('PLAY_DELAY seconds {}'.format(PLAY_DELAY))  #seconds
  print('FORMAT {}'.format(FORMAT))
  print('CHANNELS {}'.format(CHANNELS))
  print('RATE {}'.format(RATE))
  print('OUTPUT_2_WAV {}'.format(OUTPUT_2_WAV))
  print('WAV_RECIEVE_NAME {}'.format(WAV_RECIEVE_NAME))
  print('WAV_SEND_NAME {}'.format(WAV_SEND_NAME))

if __name__ == '__main__':
  if len(sys.argv) < 2:
    print('No parameters <bitrate> specified')
    print('Using defaults for 700C codec2 mode')
    CHUNK = 4480
    ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + '700C' + ' - -"'
    DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + '700C' + ' - -"'
    print_globals()
  else:
    if sys.argv[1] == '700':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 4480
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
    elif sys.argv[1] == '700C':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 4480
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
    elif sys.argv[1] == '700B':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 4480
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
    elif sys.argv[1] == '1600':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 2240
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
    elif sys.argv[1] == '1200':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 2880
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
      extra_bytes_for_coded_data = bytes([0xaa,0xaa])
      print('extra_bytes_for_coded_data {}'.format(extra_bytes_for_coded_data.hex()))
    elif sys.argv[1] == '2400':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 1440
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
      extra_bytes_for_coded_data = bytes([0xbb,0xbb])
      print('extra_bytes_for_coded_data {}'.format(extra_bytes_for_coded_data.hex()))
    elif sys.argv[1] == '3200':
      print('Parameters <bitrate> specified as <{}>'.format(sys.argv[1]))
      print('Using "{}" codec2 mode'.format(sys.argv[1]))
      CHUNK = 1120
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + sys.argv[1] + ' - -"'
      print_globals()
    else:
      print('Parameters <bitrate> specified as <{}> UNKNOWN'.format(sys.argv[1]))
      print('Using default codec2 mode')
      CHUNK = 4480
      ENC_CALL_STR = 'bash -c "' + ENC_CALL_PATH + ' ' + '700C' + ' - -"'
      DEC_CALL_STR = 'bash -c "' + DEC_CALL_PATH + ' ' + '700C' + ' - -"'
      print_globals()
  main()

#METHOD FROM codec2.c
#int codec2_bits_per_frame(struct CODEC2 *c2) {
#    if (c2->mode == CODEC2_MODE_3200)  140 ms frame in 56 bytes
#	return 64;
#    if (c2->mode == CODEC2_MODE_2400)  CHUNK 1440 -> 54 bytes | 180 ms or 54 bytes = frame 9 codec2 frames of 48 bit 
#	return 48;
#    if  (c2->mode == CODEC2_MODE_1600)
#	return 64;
#    if  (c2->mode == CODEC2_MODE_1400)
#	return 56;
#    if  (c2->mode == CODEC2_MODE_1300)
#	return 52;
#    if  (c2->mode == CODEC2_MODE_1200) CHUNK 2880 ms -> 54 bytes | 360 ms 54 bytes = frame 9 codec2 frames of 48 bit 
#	return 48;
#    if  (c2->mode == CODEC2_MODE_700)
#	return 28;
#    if  (c2->mode == CODEC2_MODE_700B)
#	return 28;
#    if  (c2->mode == CODEC2_MODE_700C)
#	return 28;
#    //TODO: verify this
#    if (c2->mode == CODEC2_MODE_WB)
#	return 64;
#    return 0; /* shouldn't get here */
#}